# -*- coding: utf-8 -*-

"""Setup and run Psi4"""

import json
import logging
from pathlib import Path

import psi4_step
import seamm
import seamm.data
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('psi4')


class Initialization(seamm.Node):

    def __init__(self, flowchart=None, title='Initialization', extension=None):
        """Initialize the node"""

        logger.debug('Creating Initialization {}'.format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.parameters = psi4_step.InitializationParameters()

        self.description = ['Initialization of Psi4']

    @property
    def header(self):
        """A printable header for this section of output"""
        return (
            'Step {}: {}'.format(
                '.'.join(str(e) for e in self._id), self.title
            )
        )

    @property
    def version(self):
        """The semantic version of this module.
        """
        return psi4_step.__version__

    @property
    def git_revision(self):
        """The git version of this module.
        """
        return psi4_step.__git_revision__

    def description_text(self, P=None):
        """Prepare information about what this node will do
        """
        if not P:
            P = self.parameters.values_to_dict()

        text = (
            f"Using the basis set '{P['basis']}' and a symmetry "
            f"tolerance of {P['symmetry_tolerance']}."
        )

        return self.header + '\n' + __(text, indent=4 * ' ').__str__()

    def get_input(self):
        """Get the input for an initialization calculation for Psi4"""
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = '{:~P}'.format(PP[key])

        self.description = []
        self.description.append(
            __(self.description_text(PP), **PP, indent=self.indent)
        )

        # Start gathering the keywords
        result = []
        result.append('')
        result.append('#' * 80)
        result.append(f'# {self.header}')
        result.append('#' * 80)

        result.append(f"set basis {P['basis']}")
        result.append('')
        result.append(f"initial.symmetrize({P['symmetry_tolerance']})")
        result.append('point_group = initial.point_group().symbol()')

        # Dump the properties to a json file
        filename = f'@{self._id[-1]}+properties.json'
        result.append('')
        result.append("variables = {'point group': point_group}")
        result.append('')
        result.append(f"with open('{filename}', 'w') as fd:")
        result.append('    json.dump(variables, fd, sort_keys=True, indent=3)')

        return '\n'.join(result)

    def analyze(self, indent='', data={}, out=[]):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        # Read in the results from json
        directory = Path(self.directory)
        json_file = directory / 'properties.json'
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)

            text = 'The calculation will use the point group {point group}.'
        else:
            data = {}
            text = (
                '\nThere are no results from Psi4. Perhaps it '
                f'failed? Looking for {str(json_file)}.'
            ),

        printer.normal(__(text, **data, indent=self.indent + 4 * ' '))



# -*- coding: utf-8 -*-

"""The graphical part of a Psi4 Initialization node"""

import configargparse
import logging
import tkinter as tk

import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkInitialization(seamm.TkNode):

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=120,
        y=20,
        w=200,
        h=50,
        my_logger=logger,
        keyword_metadata=None
    ):
        """Initialize the graphical Tk Psi4 initialization step

        Keyword arguments:
        """
        self.results_widgets = []

        # Argument/config parsing
        self.parser = configargparse.ArgParser(
            auto_env_var_prefix='',
            default_config_files=[
                '/etc/seamm/psi4_initialization.ini',
                '/etc/seamm/seamm.ini',
                '~/.seamm/psi4_initialization.ini',
                '~/.seamm/seamm.ini',
            ]
        )

        self.parser.add_argument(
            '--seamm-configfile',
            is_config_file=True,
            default=None,
            help='a configuration file to override others'
        )

        # Options for this plugin
        self.parser.add_argument(
            "--psi4-tk-initialization-log-level",
            default=configargparse.SUPPRESS,
            choices=[
                'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
            ],
            type=lambda string: string.upper(),
            help="the logging level for the Psi4 Tk_initialization step"
        )

        self.options, self.unknown = self.parser.parse_known_args()

        # Set the logging level for this module if requested
        if 'psi4_tk_initialization_log_level' in self.options:
            my_logger.setLevel(self.options.psi4_tk_initialization_log_level)
            my_logger.critical(
                'Set log level to {}'.format(
                    self.options.psi4_tk_initialization_log_level
                )
            )

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
            keyword_metadata=keyword_metadata
        )

    def right_click(self, event):
        """Probably need to add our dialog...
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self, title='Edit Psi4 Initialization Step'):
        """Create the dialog!"""
        self.logger.debug('Creating the dialog')
        frame = super().create_dialog(title=title)

        # Create all the widgets
        P = self.node.parameters
        for key in P:
            if key not in ('results', 'extra keywords', 'create tables'):
                self[key] = P[key].widget(frame)

        self.logger.debug('Finished creating the dialog')

    def reset_dialog(self, widget=None):
        frame = self['frame']
        for slave in frame.grid_slaves():
            slave.grid_forget()

        widgets = []
        row = 0
        self['basis'].grid(row=row, column=0, sticky=tk.EW)
        widgets.append(self['basis'])
        row += 1
        self['symmetry_tolerance'].grid(row=row, column=0, sticky=tk.EW)
        widgets.append(self['symmetry_tolerance'])
        row += 1
        sw.align_labels(widgets)

        return row



# -*- coding: utf-8 -*-

"""Non-graphical part of the Psi4 step in a SEAMM flowchart
"""

import json
import logging
from pathlib import Path
import pprint

import configargparse
import psutil

import psi4_step
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: 'job' and 'printer'. 'job' send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# 'printer' sends output to the file 'step.out' in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('Psi4')


def humanize(memory, suffix="B", kilo=1024):
    """
    Scale memory to its proper format
    e.g:
        1253656 => '1.20 MiB'
        1253656678 => '1.17 GiB'
    """
    if kilo == 1000:
        units = ["", "k", "M", "G", "T", "P"]
    elif kilo == 1024:
        units = ["", "Ki", "Mi", "Gi", "Ti", "Pi"]
    else:
        raise ValueError('kilo must be 1000 or 1024!')

    for unit in units:
        if memory < kilo:
            return f"{memory:.2f} {unit}{suffix}"
        memory /= kilo


def dehumanize(memory, suffix="B"):
    """
    Unscale memory from its human readable form
    e.g:
        '1.20 MB' => 1200000
        '1.17 GB' => 1170000000
    """
    units = {
        '': 1,
        'k': 1000,
        'M': 1000**2,
        'G': 1000**3,
        'P': 1000**4,
        'Ki': 1024,
        'Mi': 1024**2,
        'Gi': 1024**3,
        'Pi': 1024**4
    }

    tmp = memory.split()
    if len(tmp) == 1:
        return memory
    elif len(tmp) > 2:
        raise ValueError('Memory must be <number> <units>, e.g. 1.23 GB')

    amount, unit = tmp
    amount = float(amount)

    for prefix in units:
        if prefix + suffix == unit:
            return int(amount * units[prefix])

    raise ValueError(f"Don't recognize the units on '{memory}'")


class Psi4(seamm.Node):
    """
    The non-graphical part of a Psi4 step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : Psi4Parameters
        The control parameters for Psi4.

    See Also
    --------
    TkPsi4,
    Psi4, Psi4Parameters
    """

    def __init__(
        self,
        flowchart=None,
        title='Psi4',
        namespace='org.molssi.seamm.psi4',
        extension=None
    ):
        """A step for Psi4 in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
            flowchart: seamm.Flowchart
                The non-graphical flowchart that contains this step.

            title: str
                The name displayed in the flowchart.
            namespace: The namespace for the plugins of the subflowchart
            extension: None
                Not yet implemented
        """
        logger.debug('Creating Psi4 {}'.format(self))

        # Argument/config parsing
        self.parser = configargparse.ArgParser(
            auto_env_var_prefix='',
            default_config_files=[
                '/etcs/seamm/psi4.in',
                '/etc/seamm/psi4_step.ini',
                '/etc/seamm/seamm.ini',
                '~/.seamm/psi4.ini',
                '~/.seamm/psi4_step.ini',
                '~/.seamm/seamm.ini',
            ]
        )

        self.parser.add_argument(
            '--seamm-configfile',
            is_config_file=True,
            default=None,
            help='a configuration file to override others'
        )

        # Options for this plugin
        self.parser.add_argument(
            "--psi4-step-log-level",
            default=configargparse.SUPPRESS,
            choices=[
                'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
            ],
            type=str.upper,
            help="the logging level for the Psi4 step"
        )

        # Options for Psi4
        self.parser.add_argument(
            '--psi4-exe',
            default='psi4',
            help='the path to the Psi4 executable'
        )

        self.parser.add_argument(
            '--psi4-num-threads',
            default='default',
            help='How many threads to use in Psi4'
        )

        self.parser.add_argument(
            '--psi4-max-threads',
            default='default',
            help='The maximum number of threads to use in Psi4'
        )

        self.parser.add_argument(
            '--psi4-max-memory',
            default='default',
            help='The maximum amount of memory to use for Psi4'
        )

        self.parser.add_argument(
            '--psi4-memory',
            default='default',
            help='The amount of memory to use for Psi4'
        )

        self.parser.add_argument(
            '--psi4-memory-factor',
            default='90%',
            help=(
                'The amount of possible memory to use, typically about 90% to '
                'allow for Psi4 not keeping track of all the memory used.'
            )
        )

        # General options from seamm.ini on max cores and memory
        self.parser.add_argument(
            '--max-cores',
            default='unlimited',
            help='The maximum number of cores/threads to use in any step.'
        )

        self.parser.add_argument(
            '--max-memory',
            default='available',
            help='The maximum amount of memory to use in any step.'
        )

        self.options, self.unknown = self.parser.parse_known_args()

        # Set the logging level for this module if requested
        if 'psi4_step_log_level' in self.options:
            logger.setLevel(self.options.psi4_step_log_level)
        self.subflowchart = seamm.Flowchart(
            parent=self, name='Psi4', namespace=namespace
        )

        super().__init__(
            flowchart=flowchart, title='Psi4', extension=extension
        )

        self.parameters = psi4_step.Psi4Parameters()

    @property
    def version(self):
        """The semantic version of this module.
        """
        return psi4_step.__version__

    @property
    def git_revision(self):
        """The git version of this module.
        """
        return psi4_step.__git_revision__

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        self._id = node_id

        # and set our subnodes
        self.subflowchart.set_ids(self._id)

        return self.next()

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
            P: dict
                An optional dictionary of the current values of the control
                parameters.
        Returns
        -------
            description : str
                A description of the current step.
        """

        self.subflowchart.root_directory = self.flowchart.root_directory

        # Get the first real node
        node = self.subflowchart.get_node('1').next()

        text = self.header + '\n\n'
        while node is not None:
            text += __(node.description_text(), indent=3 * ' ').__str__()
            text += '\n'
            node = node.next()

        return text

    def run(self):
        """Run a Psi4 step.

        Returns
        -------

        next_node : seamm.Node
            The next node object in the flowchart.

        """
        printer.important(self.header)
        printer.important('')

        system = self.get_variable('_system')
        n_atoms = system.n_atoms()
        if n_atoms == 0:
            logger.error('Psi4 run(): there is no structure!')
            raise RuntimeError('Psi4 run(): there is no structure!')

        # Access the options
        o = self.options

        # How many processors does this node have?
        n_cores = psutil.cpu_count(logical=False)
        logger.info('The number of cores is {}'.format(n_cores))

        # How many threads to use
        if o.psi4_num_threads == 'default':
            n_threads = n_cores
        else:
            n_threads = int(o.psi4_num_threads)
        if n_threads > n_cores:
            n_threads = n_cores
        if n_threads < 1:
            n_threads = 1
        if o.psi4_max_threads != 'default':
            if n_threads > int(o.psi4_max_threads):
                n_threads = int(o.psi4_max_threads)
        if o.max_cores != "unlimited":
            if n_threads > int(o.max_cores):
                n_threads = int(o.max_cores)
        logger.info(f'Psi4 will use {n_threads} threads.')

        # How much memory to use
        if '%' in o.psi4_memory_factor:
            factor = float(o.psi4_memory_factor.rstrip('%')) / 100.0
        else:
            factor = float(o.psi4_memory_factor)
        svmem = psutil.virtual_memory()
        if o.max_memory == 'all':
            available = svmem.total * factor
        else:
            available = svmem.available * factor
        if o.psi4_memory == 'default':
            memory = available * (n_threads / n_cores)
        else:
            memory = dehumanize(o.psi4_memory)
        if o.psi4_max_memory != 'default':
            max_memory = dehumanize(o.psi4_max_memory)
            if memory > max_memory:
                memory = max_memory
        if o.max_memory != 'all' and o.max_memory != 'available':
            if memory > dehumanize(o.max_memory):
                memory = dehumanize(o.max_memory)
        # Psi applies a minimum of 250 MiB
        min_memory = dehumanize('250 MiB')
        if min_memory > memory:
            memory = min_memory
        memory = humanize(memory, kilo=1024)

        printer.important(
            self.indent +
            f'    Psi4 will use {n_threads} threads and {memory} memory\n'
        )

        # Start the input data
        input_data = []
        input_data.append('import json')
        input_data.append('import numpy as np')
        input_data.append('')
        input_data.append(f'memory {memory}')
        input_data.append('')

        # Work through the subflowchart to find out what to do.
        self.subflowchart.root_directory = self.flowchart.root_directory

        next_node = super().run(printer)

        # Get the first real node
        node = self.subflowchart.get_node('1').next()

        # Put the structure into the input
        input_data.append(self._convert_structure(name='initial'))

        while node is not None:
            text = node.get_input()
            input_data.append(text)
            node = node.next()

        # Write out the final structure
        input_data.append('')
        input_data.append('# Write the final structure to disk')
        input_data.append('molecule = get_active_molecule()')
        input_data.append('tmp = molecule.to_dict()')
        input_data.append('for item, value in tmp.items():')
        input_data.append('    if isinstance(value, np.ndarray):')
        input_data.append('        tmp[item] = value.tolist()')
        input_data.append('')
        input_data.append("with open('final_structure.json', 'w') as fd:")
        input_data.append('    json.dump(tmp, fd, sort_keys=True, indent=3)')

        files = {'input.dat': '\n'.join(input_data)}
        logger.info('input.dat:\n' + files['input.dat'])

        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)
        for filename in files:
            path = directory / filename
            with path.open(mode='w') as fd:
                fd.write(files[filename])

        return_files = ['output.dat', '*properties.json', '*structure.json']
        env = {'PSIPATH': Path(o.psi4_exe).parent}

        local = seamm.ExecLocal()
        result = local.run(
            cmd=[o.psi4_exe, f'-n {n_threads}'],
            files=files,
            return_files=return_files,
            env=env
        )  # yapf: disable

        if result is None:
            logger.error('There was an error running Psi4')
            return None

        logger.debug('\n' + pprint.pformat(result))

        logger.debug('stdout:\n' + result['stdout'])
        if 'stdout' in result and result['stdout'] != '':
            path = directory / 'stdout.txt'
            with path.open(mode='w') as fd:
                fd.write(result['stdout'])

        if result['stderr'] != '':
            logger.warning('stderr:\n' + result['stderr'])
            path = directory / 'stderr.txt'
            with path.open(mode='w') as fd:
                fd.write(result['stderr'])

        for filename in result['files']:
            if filename[0] == '@':
                subdir, fname = filename[1:].split('+')
                path = directory / subdir / fname
            else:
                path = directory / filename
            with path.open(mode='w') as fd:
                if result[filename]['data'] is not None:
                    fd.write(result[filename]['data'])
                else:
                    fd.write(result[filename]['exception'])

        # Analyze the results
        self.analyze()

        return next_node

    def analyze(self, indent='', **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        'printer'.

        Parameters
        ----------
            indent: str
                An extra indentation for the output
        """

        # Get the first real node
        node = self.subflowchart.get_node('1').next()

        # Loop over the subnodes, asking them to do their analysis
        while node is not None:
            for value in node.description:
                printer.important(value)

            node.analyze()

            printer.normal('')

            node = node.next()

        # Update the structure
        directory = Path(self.directory)
        structure_file = directory / 'final_structure.json'
        if structure_file.exists():
            with structure_file.open(mode='r') as fd:
                structure = json.load(fd)
        if 'geom' in structure:
            system = self.get_variable('_system')
            xs = []
            ys = []
            zs = []
            it = iter(structure['geom'])
            for x in it:
                xs.append(x)
                ys.append(next(it))
                zs.append(next(it))
            system.atoms['x'][0:] = xs
            system.atoms['y'][0:] = ys
            system.atoms['z'][0:] = zs
            printer.important(
                self.indent +
                '    Updated the system with the structure from Psi4',
            )
            printer.important('')

    def _convert_structure(self, name=None):
        """Convert the structure to the input for Psi4."""

        system = self.get_variable('_system')

        structure = []
        if name is None:
            structure.append('molecule {')
        else:
            structure.append('molecule ' + name + ' {')

        # Charge and multiplicity
        if 'extras' in system:
            extras = system['extras']

            if 'open' in extras and extras['open'] is not None:
                openshell = extras['open']
                if (
                    (
                        isinstance(openshell, tuple) or
                        isinstance(openshell, list)
                    ) and len(openshell) > 1
                ):
                    nopen = openshell[0]
                    norbitals = openshell[1]
                    if nopen != norbitals:
                        raise NotImplementedError(
                            f"Handling of open shell = '{openshell}'"
                        )
                else:
                    nopen = openshell
                    norbitals = nopen

                if 'net_charge' in extras and extras['net_charge'] is not None:
                    structure.append(f"    {extras['net_charge']}   {nopen}")
                else:
                    structure.append(f'    0   {nopen}')
            else:
                if 'net_charge' in extras and extras['net_charge'] is not None:
                    structure.append(f"    {extras['net_charge']}   1")

        elements = system.atoms.symbols()
        coordinates = system.atoms.coordinates()

        if 'freeze' in system.atoms:
            freeze = system.atoms['freeze']
        else:
            freeze = [''] * len(elements)

        for element, xyz, frz in zip(elements, coordinates, freeze):
            x, y, z = xyz
            structure.append(
                f'    {element:2s} {float(x): 12.8f} {float(y): 12.8f} '
                f'{float(z): 12.8f}'
            )
        structure.append('}')

        return '\n'.join(structure) + '\n'



# -*- coding: utf-8 -*-

"""The graphical part of a Psi4 Optimization node"""

import configargparse
import logging
import pprint
import tkinter as tk
import tkinter.ttk as ttk

import psi4_step
# import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkOptimization(psi4_step.TkEnergy):

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=120,
        y=20,
        w=200,
        h=50,
        my_logger=logger,
        keyword_metadata=None
    ):
        """Initialize the graphical Tk Psi4 optimization step

        Keyword arguments:
        """
        self.results_widgets = []

        # Argument/config parsing
        self.parser = configargparse.ArgParser(
            auto_env_var_prefix='',
            default_config_files=[
                '/etc/seamm/psi4_optimization.ini',
                '/etc/seamm/seamm.ini',
                '~/.seamm/psi4_optimization.ini',
                '~/.seamm/seamm.ini',
            ]
        )

        self.parser.add_argument(
            '--seamm-configfile',
            is_config_file=True,
            default=None,
            help='a configuration file to override others'
        )

        # Options for this plugin
        self.parser.add_argument(
            "--psi4-tk-optimization-log-level",
            default=configargparse.SUPPRESS,
            choices=[
                'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
            ],
            type=lambda string: string.upper(),
            help="the logging level for the Psi4 Tk_optimization step"
        )

        self.options, self.unknown = self.parser.parse_known_args()

        # Set the logging level for this module if requested
        if 'psi4_tk_optimization_log_level' in self.options:
            my_logger.setLevel(self.options.psi4_tk_optimization_log_level)
            my_logger.critical(
                'Set log level to {}'.format(
                    self.options.psi4_tk_optimization_log_level
                )
            )

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
            keyword_metadata=keyword_metadata
        )

    def right_click(self, event):
        """Probably need to add our dialog...
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(
        self, title='Edit Psi4 Optimization Step', calculation='optimization'
    ):
        """Create the edit dialog!

        This is reasonably complicated, so a bit of description
        is in order. The superclass Energy creates the dialog
        along with the calculation parameters in a 'calculation'
        frame..

        This method adds a second frame for controlling the optimizer.

        The layout is handled in part by the Energy superclass, which
        handles the calculation frame. Our part is handled by two
        methods:

        * reset_dialog does the general layout of the main frames.
        * reset_optimization handles the layout of the optimization
          section.
        """

        logger.debug('TkOptimization.create_dialog')

        # Let parent classes do their thing.
        super().create_dialog(title=title, calculation=calculation)

        # Shortcut for parameters
        P = self.node.parameters

        logger.debug('Parameters:\n{}'.format(pprint.pformat(P.to_dict())))

        # Frame to isolate widgets
        opt_frame = self['optimization'] = ttk.LabelFrame(
            self['frame'],
            borderwidth=4,
            relief='sunken',
            text='Geometry Optimization',
            labelanchor='n',
            padding=10
        )

        for key in psi4_step.OptimizationParameters.parameters:
            self[key] = P[key].widget(opt_frame)

        # and binding to change as needed
        self['geometry convergence'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_optimization
        )
        self['geometry convergence'].combobox.bind(
            "<Return>", self.reset_optimization
        )
        self['geometry convergence'].combobox.bind(
            "<FocusOut>", self.reset_optimization
        )

        # Top level needs to call reset_dialog
        if calculation == 'optimization':
            self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets, letting our parents go first."""
        row = super().reset_dialog()

        self['optimization'].grid(row=row, column=0)
        row += 1

        self.reset_optimization()

        return row

    def reset_optimization(self, widget=None):
        convergence = self['geometry convergence'].get()

        frame = self['optimization']
        for slave in frame.grid_slaves():
            slave.grid_forget()

        widgets = []
        # widgets2 = []
        row = 0

        self['optimization method'].grid(
            row=row, column=0, columnspan=2, sticky=tk.EW
        )
        widgets.append(self['optimization method'])
        row += 1

        self['max geometry steps'].grid(
            row=row, column=0, columnspan=2, sticky=tk.EW
        )
        widgets.append(self['max geometry steps'])
        row += 1

        self['geometry convergence'].grid(
            row=row, column=0, columnspan=2, sticky=tk.EW
        )
        widgets.append(self['geometry convergence'])
        row += 1

        if convergence == 'Custom':
            pass

        self['recalc hessian'].grid(
            row=row, column=0, columnspan=2, sticky=tk.EW
        )
        widgets.append(self['recalc hessian'])
        row += 1

        sw.align_labels(widgets)



methods = {
    'effective fragment potential (EFP)': {
        'method': 'efp',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic'
    },
    'Kohn-Sham (KS) density functional theory (DFT)': {
        'method': 'dft',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic'
    },
    'Hartree-Fock (HF) self consistent field (SCF)': {
        'method': 'hf',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic'
    },
    'HF with dispersion, BSSE, and basis set corrections': {
        'method': 'hf3c',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference'
    },
    'PBEh with dispersion, BSSE, and basis set corrections': {
        'method': 'pbeh3c',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference'
    },
    'density cumulant functional theory': {
        'method': 'dcft',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic'
    },
    '2nd-order Møller–Plesset perturbation theory (MP2)': {
        'method': 'mp2',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic',
        'freeze core?': True
    },
    '3rd-order Møller–Plesset perturbation theory (MP3)': {
        'method': 'mp3',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'MP3 with frozen natural orbitals': {
        'method': 'fno-mp3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'average of MP2 and MP3': {
        'method': 'mp2.5',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic',
        'freeze core?': True
    },
    '4th-order MP perturbation theory (MP4) less triples': {
        'method': 'mp4(sdq)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'MP4 (less triples) with frozen natural orbitals': {
        'method': 'fno-mp4(sdq)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'full MP4': {
        'method': 'mp4',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'full MP4 with frozen natural orbitals': {
        'method': 'fno-mp4',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'nth-order Møller–Plesset (MP) perturbation theory': {
        'method': 'mpn',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'nth-order z-averaged perturbation theory (ZAPT)': {
        'method': 'zaptn',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'orbital-optimized second-order MP perturbation theory': {
        'method': 'omp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'spin-component scaled OMP2': {
        'method': 'scs-omp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'a special version of SCS-OMP2 for nucleobase interactions': {
        'method': 'scs(n)-omp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'a special version of SCS-OMP2 (from ethene dimers)': {
        'method': 'scs-omp2-vdw',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'spin-opposite scaled OMP2': {
        'method': 'sos-omp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'A special version of SOS-OMP2 for pi systems': {
        'method': 'sos-pi-omp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'orbital-optimized third-order MP perturbation theory': {
        'method': 'omp3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'spin-component scaled OMP3': {
        'method': 'scs-omp3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'a special version of SCS-OMP3 for nucleobase interactions': {
        'method': 'scs(n)-omp3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'a special version of SCS-OMP3 (from ethene dimers)': {
        'method': 'scs-omp3-vdw',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'spin-opposite scaled OMP3': {
        'method': 'sos-omp3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'A special version of SOS-OMP3 for pi systems': {
        'method': 'sos-pi-omp3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
        },
    'orbital-optimized MP2.5': {
        'method': 'omp2.5',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'coupled electron pair approximation variant 0': {
        'method': 'cepa(0)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CEPA(0) with frozen natural orbitals': {
        'method': 'fno-cepa(0)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'coupled electron pair approximation variant 1': {
        'method': 'cepa(1)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CEPA(1) with frozen natural orbitals': {
        'method': 'fno-cepa(1)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'coupled electron pair approximation variant 3': {
        'method': 'cepa(3)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CEPA(3) with frozen natural orbitals': {
        'method': 'fno-cepa(3)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'averaged coupled-pair functional': {
        'method': 'acpf',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'ACPF with frozen natural orbitals': {
        'method': 'fno-acpf',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'averaged quadratic coupled cluster': {
        'method': 'aqcc',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'AQCC with frozen natural orbitals': {
        'method': 'fno-aqcc',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'quadratic CI singles doubles (QCISD)': {
        'method': 'qcisd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'QCISD with frozen natural orbitals': {
        'method': 'fno-qcisd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'Linear CCD': {
        'method': 'lccd',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'LCCD with frozen natural orbitals': {
        'method': 'fno-lccd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'orbital optimized LCCD': {
        'method': 'olccd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'approximate coupled cluster singles and doubles (CC2)': {
        'method': 'cc2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'coupled cluster doubles (CCD)': {
        'method': 'ccd',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'coupled cluster singles and doubles (CCSD)': {
        'method': 'ccsd',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'Brueckner coupled cluster doubles (BCCD)': {
        'method': 'bccd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CCSD with frozen natural orbitals': {
        'method': 'fno-ccsd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'QCISD with perturbative triples': {
        'method': 'qcisd(t)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'QCISD(T) with frozen natural orbitals': {
        'method': 'fno-qcisd(t)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CCSD with perturbative triples (CCSD(T))': {
        'method': 'ccsd(t)',
        'calculation': ['energy', 'gradients'],
        'level': 'normal',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'CCSD with asymmetric perturbative triples (CCSD(AT))': {
        'method': 'ccsd(at)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'BCCD with perturbative triples': {
        'method': 'bccd(t)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CCSD(T) with frozen natural orbitals': {
        'method': 'fno-ccsd(t)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'approximate CC singles, doubles, and triples (CC3)': {
        'method': 'cc3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'expert full control over ccenergy module': {
        'method': 'ccenergy',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'expert full control over dfocc module': {
        'method': 'dfocc',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'configuration interaction (CI) singles and doubles (CISD)': {
        'method': 'cisd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CISD with frozen natural orbitals': {
        'method': 'fno-cisd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CI singles, doubles, and triples (CISDT)': {
        'method': 'cisdt',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'CI singles, doubles, triples, and quadruples (CISDTQ)': {
        'method': 'cisdtq',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'nth-order CI': {
        'method': 'cin',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'full configuration interaction (FCI)': {
        'method': 'fci',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'expert full control over detci module': {
        'method': 'detci',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'complete active space self consistent field (CASSCF)': {
        'method': 'casscf',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': False
    },
    'restricted active space self consistent field (RASSCF)': {
        'method': 'rasscf',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': False
    },
    'multiconfigurational self consistent field (SCF)': {
        'method': 'mcscf',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': False
    },
    'Mukherjee multireference coupled cluster (Mk-MRCC)': {
        'method': 'psimrcc',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': False
    },
    'density matrix renormalization group SCF': {
        'method': 'dmrg-scf',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': False
    },
    'density matrix renormalization group CASPT2': {
        'method': 'dmrg-caspt2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'density matrix renormalization group CI': {
        'method': 'dmrg-ci',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
        },
    '0th-order symmetry adapted perturbation theory (SAPT)': {
        'method': 'sapt0',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    '0th-order SAPT with special exchange scaling': {
        'method': 'ssapt0',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    '0th-order functional and/or intramolecular SAPT': {
        'method': 'fisapt0',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    '2nd-order SAPT, traditional definition': {
        'method': 'sapt2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT including all 2nd-order terms': {
        'method': 'sapt2+',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT including perturbative triples': {
        'method': 'sapt2+(3)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT including all 3rd-order terms': {
        'method': 'sapt2+3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+ with CC-based dispersion': {
        'method': 'sapt2+(ccd)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+(3) with CC-based dispersion': {
        'method': 'sapt2+(3)(ccd)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+3 with CC-based dispersion': {
        'method': 'sapt2+3(ccd)',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT including all 2nd-order terms and MP2 correction': {
        'method': 'sapt2+dmp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT including perturbative triples and MP2 correction': {
        'method': 'sapt2+(3)dmp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT including all 3rd-order terms and MP2 correction': {
        'method': 'sapt2+3dmp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+ with CC-based dispersion and MP2 correction': {
        'method': 'sapt2+(ccd)dmp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+(3) with CC-based dispersion and MP2 correction': {
        'method': 'sapt2+(3)(ccd)dmp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+3 with CC-based dispersion and MP2 correction': {
        'method': 'sapt2+3(ccd)dmp2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    '0th-order SAPT plus charge transfer (CT) calculation': {
        'method': 'sapt0-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2 plus CT': {
        'method': 'sapt2-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+ plus CT': {
        'method': 'sapt2+-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+(3) plus CT': {
        'method': 'sapt2+(3)-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+3 plus CT': {
        'method': 'sapt2+3-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+(CCD) plus CT': {
        'method': 'sapt2+(ccd)-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+(3)(CCD) plus CT': {
        'method': 'sapt2+(3)(ccd)-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'SAPT2+3(CCD) plus CT': {
        'method': 'sapt2+3(ccd)-ct',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    '2nd-order algebraic diagrammatic construction (ADC)': {
        'method': 'adc',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'EOM-CC2': {
        'method': 'eom-cc2',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    },
    'equation of motion (EOM) CCSD': {
        'method': 'eom-ccsd',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'analytic',
        'freeze core?': True
    },
    'EOM-CC3': {
        'method': 'eom-cc3',
        'calculation': ['energy', 'gradients'],
        'level': 'expert',
        'gradients': 'finite-difference',
        'freeze core?': True
    }
}  # yapf: disable

dft_functionals = {
    'B1LYP Hyb-GGA Exchange-Correlation Functional': {
        'method': 'b1lyp',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'B1PW91 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b1pw91',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B1WC Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b1wc',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B2GPPLYP Double Hybrid Exchange-Correlation Functional': {
        'name': 'b2gpplyp',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'B2PLYP Double Hybrid Exchange-Correlation Functional': {
        'name': 'b2plyp',
        'dispersion': ['none', 'd3bj', 'd3mbj', 'nl'],
        'level': 'expert',
    },
    'B3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b3lyp',
        'dispersion': ['none', 'd3bj', 'd3mbj', 'nl'],
        'level': 'normal',
    },
    'B3LYP5 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b3lyp5',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B3LYPs Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b3lyps',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B3P86 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b3p86',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'B3PW91 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b3pw91',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'B5050LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b5050lyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B86B95 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b86b95',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B86BPBE GGA Exchange-Correlation Functional': {
        'name': 'b86bpbe',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B88B95 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b88b95',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'B97-0 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b97-0',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B97-1 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b97-1',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'B97-1p Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b97-1p',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B97-2 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b97-2',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'B97-3 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b97-3',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'b97-d': {
        'name': 'b97-d',
        'dispersion': ['none', 'd3bj', 'd3mbj'],
        'level': 'expert',
    },
    'B97-GGA1 GGA Exchange-Correlation Functional': {
        'name': 'b97-gga1',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B97-K Hyb-GGA Exchange-Correlation Functional': {
        'name': 'b97-k',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'b97m-d3bj': {
        'name': 'b97m-d3bj',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'B97M-V GGA Exchange-Correlation Functional': {
        'name': 'b97m-v',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'BB1K Hyb-GGA Exchange-Correlation Functional': {
        'name': 'bb1k',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'BHandH Hyb-GGA Exchange-Correlation Functional': {
        'name': 'bhandh',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'BHandHLYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'bhandhlyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'BLYP GGA Exchange-Correlation Functional': {
        'name': 'blyp',
        'dispersion': ['none', 'd3bj', 'd3mbj', 'nl'],
        'level': 'expert',
    },
    'BOP GGA Exchange-Correlation Functional': {
        'name': 'bop',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'BP86 GGA Exchange-Correlation Functional': {
        'name': 'bp86',
        'dispersion': ['none', 'd3bj', 'd3mbj', 'nl'],
        'level': 'expert',
    },
    'CAM-B3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'cam-b3lyp',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'CAP0 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'cap0',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'core-dsd-blyp': {
        'name': 'core-dsd-blyp',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'Dispersionless Hybrid Meta-GGA XC Functional': {
        'name': 'dldf',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Dispersionless Hybrid Meta-GGA XC Functional + d09': {
        'name': 'dldf+d09',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Dispersionless Hybrid Meta-GGA XC Functional + d10': {
        'name': 'dldf+d10',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'DSD-BLYP SCS Double Hybrid XC Functional': {
        'name': 'dsd-blyp',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'DSD-PBEB95 SCS Double Hybrid Meta-GGA XC Functional': {
        'name': 'dsd-pbeb95',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'DSD-PBEP86 SCS Double Hybrid XC Functional': {
        'name': 'dsd-pbep86',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'DSD-PBEPBE SCS Double Hybrid XC Functional': {
        'name': 'dsd-pbepbe',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'EDF1 GGA Exchange-Correlation Functional': {
        'name': 'edf1',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'EDF2 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'edf2',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'FT97 GGA Exchange-Correlation Functional': {
        'name': 'ft97',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'GAM GGA Minnesota Exchange-Correlation Functional': {
        'name': 'gam',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HCTH120 GGA Exchange-Correlation Functional': {
        'name': 'hcth120',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'HCTH147 GGA Exchange-Correlation Functional': {
        'name': 'hcth147',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HCTH407 GGA Exchange-Correlation Functional': {
        'name': 'hcth407',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'HCTH407P GGA Exchange-Correlation Functional': {
        'name': 'hcth407p',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HCTH93 GGA Exchange-Correlation Functional': {
        'name': 'hcth93',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HCTHP14 GGA Exchange-Correlation Functional': {
        'name': 'hcthp14',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HCTHP76 GGA Exchange-Correlation Functional': {
        'name': 'hcthp76',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'hf': {
        'name': 'hf',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'hf+d': {
        'name': 'hf+d ',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Hartree Fock based 3C composite method with minimal basis set, gCP and D3(BJ)': {  # noqa: E501
        'name': 'hf3c',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HJS-B88 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hjs-b88',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HJS-B97X Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hjs-b97x',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HJS-PBE Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hjs-pbe',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HJS-PBE-SOL Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hjs-pbe-sol',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HPBEINT Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hpbeint',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'HSE03 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hse03',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'HSE06 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'hse06',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'KSDT Exchange-Correlation Functional': {
        'name': 'ksdt',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'KT2 GGA Exchange-Correlation Functional': {
        'name': 'kt2',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'LC-VV10 GGA Exchange-Correlation Functional': {
        'name': 'lc-vv10',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'LRC-WPBE GGA Exchange-Correlation Functional': {
        'name': 'lrc-wpbe',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'LRC-wPBEh Hyb-GGA Exchange-Correlation Functional': {
        'name': 'lrc-wpbeh',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'M05 Meta-GGA XC Functional': {
        'name': 'm05',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Heavily Parameterized Hybrid M05-2X Meta-GGA XC Functional': {
        'name': 'm05-2x',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'M06 Meta-GGA XC Functional': {
        'name': 'm06',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Hybrid M06-2X Meta-GGA XC Functional': {
        'name': 'm06-2x',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Minnesota M06-HF Hybrid XC Functional': {
        'name': 'm06-hf',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'M06-L Meta-GGA XC Functional': {
        'name': 'm06-l',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Minnesota M08-HX Hybrid XC Functional': {
        'name': 'm08-hx',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'Minnesota M08-SO Hybrid XC Functional': {
        'name': 'm08-so',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'M11 Meta-GGA XC Functional': {
        'name': 'm11',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'M11-L Meta-GGA XC Functional': {
        'name': 'm11-l',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'MB3LYP-RC04 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mb3lyp-rc04',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MGGA_MS0 Meta-GGA XC Functional': {
        'name': 'mgga_ms0',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MGGA_MS1 Meta-GGA XC Functional': {
        'name': 'mgga_ms1',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MGGA_MS2 Meta-GGA XC Functional': {
        'name': 'mgga_ms2',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MGGA_MS2h Hybrid Meta-GGA XC Functional': {
        'name': 'mgga_ms2h',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MGGA_MVS Meta-GGA XC Functional': {
        'name': 'mgga_mvs',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MGGA_MV2h Hybrid Meta-GGA XC Functional': {
        'name': 'mgga_mvsh',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MN12-L Meta-GGA XC Functional': {
        'name': 'mn12-l',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'MN12-SX Meta-GGA Hybrid Screened Exchange-Correlation Functional': {
        'name': 'mn12-sx',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'MN15 Hybrid Meta-GGA Exchange-Correlation Functional': {
        'name': 'mn15',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'MN15-L Meta-GGA XC Functional': {
        'name': 'mn15-l',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MOHLYP GGA Exchange-Correlation Functional': {
        'name': 'mohlyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'MOHLYP2 GGA Exchange-Correlation Functional': {
        'name': 'mohlyp2',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPW1B95 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpw1b95',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'mPW1K Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpw1k',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPW1LYP Hybrid GGA Exchange-Correlation Functional': {
        'name': 'mpw1lyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPW1PBE Hybrid GGA Exchange-Correlation Functional': {
        'name': 'mpw1pbe',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPW1PW Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpw1pw',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'mPW3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpw3lyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPW3PW Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpw3pw',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPWB1K Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpwb1k',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'mPWLYP1M Hyb-GGA Exchange-Correlation Functional': {
        'name': 'mpwlyp1m',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPWLYP1W GGA Exchange-Correlation Functional': {
        'name': 'mpwlyp1w',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'mPWPW GGA Exchange-Correlation Functional': {
        'name': 'mpwpw',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'N12 nonseparable GGA Exchange-Correlation Functional': {
        'name': 'n12',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'N12-SX Hybrid nonseparable GGA Exchange-Correlation Functional': {
        'name': 'n12-sx',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'O3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'o3lyp',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'oblyp-e': {
        'name': 'oblyp-d',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE GGA Exchange, One-parameter Progressive correlation Functional': {
        'name': 'op-pbe',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'opbe-d': {
        'name': 'opbe-d',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'opwlyp-d': {
        'name': 'opwlyp-d',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'otpss-d': {
        'name': 'otpss-d',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE GGA Exchange-Correlation Functional': {
        'name': 'pbe',
        'dispersion': ['none', 'd3bj', 'd3mbj', 'nl'],
        'level': 'normal',
    },
    'PBE0 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'pbe0',
        'dispersion': ['none', 'd3bj', 'd3mbj', 'nl'],
        'level': 'normal',
    },
    'PBE0-13 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'pbe0-13',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE0-2 Double Hybrid Exchange-Correlation Functional': {
        'name': 'pbe0-2',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'pbe0-dh': {
        'name': 'pbe0-dh',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE1W GGA Exchange-Correlation Functional': {
        'name': 'pbe1w',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE50 Hybrid GGA Exchange-Correlation Functional': {
        'name': 'pbe50',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE Hybrid based 3C composite method with a small basis set, gCP and D3(BJ)': {  # noqa: E501
        'name': 'pbeh3c',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBELYP1W GGA Exchange-Correlation Functional': {
        'name': 'pbelyp1w',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PKZB Meta-GGA XC Functional': {
        'name': 'pkzb',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PTPSS SOS Double Hybrid XC Functional': {
        'name': 'ptpss',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'PW6B95 Hybrid Meta-GGA XC Functional': {
        'name': 'pw6b95',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'PW86B95 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'pw86b95',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PW86PBE GGA Exchange-Correlation Functional': {
        'name': 'pw86pbe',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PW91 GGA Exchange-Correlation Functional': {
        'name': 'pw91',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'PWB6K Hyb-GGA Exchange-Correlation Functional': {
        'name': 'pwb6k',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'PWPB95 SOS Double Hybrid XC Functional': {
        'name': 'pwpb95',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'revB3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'revb3lyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'revPBE GGA Exchange-Correlation Functional': {
        'name': 'revpbe',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'revPBE0 Hybrid GGA Exchange-Correlation Functional': {
        'name': 'revpbe0',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'revised TPSS Meta-GGA XC Functional': {
        'name': 'revtpss',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'revTPSSh Hyb-GGA Exchange-Correlation Functional': {
        'name': 'revtpssh',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'RPBE GGA Exchange-Correlation Functional': {
        'name': 'rpbe',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'SB98-1a Hyb-GGA Exchange-Correlation Functional': {
        'name': 'sb98-1a',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SB98-1b Hyb-GGA Exchange-Correlation Functional': {
        'name': 'sb98-1b',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SB98-1c Hyb-GGA Exchange-Correlation Functional': {
        'name': 'sb98-1c',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SB98-2a Hyb-GGA Exchange-Correlation Functional': {
        'name': 'sb98-2a',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SB98-2b Hyb-GGA Exchange-Correlation Functional': {
        'name': 'sb98-2b',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SB98-2c Hyb-GGA Exchange-Correlation Functional': {
        'name': 'sb98-2c',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SOGGA Exchange + PBE Correlation Functional': {
        'name': 'sogga',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SOGGA11 Exchange-Correlation Functional': {
        'name': 'sogga11',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'SOGGA11-X Hybrid Exchange-Correlation Functional': {
        'name': 'sogga11-x',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'SVWN3 (RPA) LSDA Functional': {
        'name': 'svwn',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TETER93 Exchange-Correlation Functional': {
        'name': 'teter93',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH-FC GGA Exchange-Correlation Functional': {
        'name': 'th-fc',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH-FCFO GGA Exchange-Correlation Functional': {
        'name': 'th-fcfo',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH-FCO GGA Exchange-Correlation Functional': {
        'name': 'th-fco',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH-FL GGA Exchange-Correlation Functional': {
        'name': 'th-fl',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH1 GGA Exchange-Correlation Functional': {
        'name': 'th1',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH2 GGA Exchange-Correlation Functional': {
        'name': 'th2',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH3 GGA Exchange-Correlation Functional': {
        'name': 'th3',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TH4 GGA Exchange-Correlation Functional': {
        'name': 'th4',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TPSS Meta-GGA XC Functional': {
        'name': 'tpss',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'TPSSh Hyb-GGA Exchange-Correlation Functional': {
        'name': 'tpssh',
        'dispersion': ['none', 'd3bj', 'nl'],
        'level': 'expert',
    },
    'TPSSLYP1W GGA Exchange-Correlation Functional': {
        'name': 'tpsslyp1w',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'TUNED-CAM-B3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'tuned-cam-b3lyp',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'VSXC Meta-GGA XC Functional': {
        'name': 'vsxc',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'VV10 GGA Exchange-Correlation Functional': {
        'name': 'vv10',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'wB97 GGA Exchange-Correlation Functional': {
        'name': 'wb97',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'wB97M-V Hyb-GGA Exchange-Correlation Functional': {
        'name': 'wb97m-v',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'wB97X Hyb-GGA Exchange-Correlation Functional': {
        'name': 'wb97x',
        'dispersion': ['none', 'd3bj', 'd'],
        'level': 'expert',
    },
    'wB97X-V Hyb-GGA Exchange-Correlation Functional': {
        'name': 'wb97x-v',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'PBE SR-XC Functional (HJS Model)': {
        'name': 'wpbe',
        'dispersion': ['none', 'd3bj', 'd3mbj'],
        'level': 'expert',
    },
    'PBE0 SR-XC Functional (HJS Model)': {
        'name': 'wpbe0',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'X1B95 Hyb-GGA Exchange-Correlation Functional': {
        'name': 'x1b95',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'X3LYP Hyb-GGA Exchange-Correlation Functional': {
        'name': 'x3lyp',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'XB1K Hyb-GGA Exchange-Correlation Functional': {
        'name': 'xb1k',
        'dispersion': ['none'],
        'level': 'expert',
    },
    'XLYP GGA Exchange-Correlation Functional': {
        'name': 'xlyp',
        'dispersion': ['none', 'd3bj'],
        'level': 'expert',
    },
    'ZLP GGA Exchange-Correlation Functional': {
        'name': 'zlp',
        'dispersion': ['none'],
        'level': 'expert',
    },
}  # yapf: disable

optimization_methods = {
    'Rational Function Optimization': 'RFO',
    'Newton-Raphson': 'NR',
    'Steepest Descents': 'SD',
    'Static Linesearch': 'LINESEARCH_STATIC'
}

optimization_convergence = {
    'CFOUR': 'CFOUR',
    'Gaussian loose': 'GAU_LOOSE',
    'Gaussian': 'GAU',
    'Gaussian tight': 'GAU_TIGHT',
    'Gaussian very tight': 'GAU_VERYTIGHT',
    'MOLPRO': 'MOLPRO',
    'NWChem loose': 'NWCHEM_LOOSE',
    'NWChem': 'GAU',
    'NWChem tight': 'GUA_TIGHT',
    'QChem': 'QCHEM',
    'TURBOMOLE': 'TURBOMOLE',
    'Interfragment tight': 'INTERFRAG_TIGHT',
    'Custom': '',
}

properties = {
    "-D ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "dispersion correction energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Ha"
    },
    "32-POLE XXXXX": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXXXY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXXXZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXXYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXXYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXXZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXYYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXYYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXYZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XXZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XYYYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XYYYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XYYZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XYZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE XZZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE YYYYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE YYYYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE YYYZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE YYZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE YZZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "32-POLE ZZZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "5th order electrical multipole",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DFT FUNCTIONAL TOTAL ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "total energy of DFT functional",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DFT TOTAL ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "Total DFT energy including dispersion",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DFT VV10 ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "VV10 energy in DFT",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DFT XC ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "DFT exchange-correlation energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DIPOLE X": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical dipole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DIPOLE Y": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical dipole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DIPOLE Z": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical dipole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "DISPERSION CORRECTION ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "energy of the dispersion correction",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "ESP": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrostatic potential at nuclei",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XXXX": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XXXY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XXXZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XXYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XXYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XXZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XYYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XYYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XYZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE XZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE YYYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE YYYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE YYZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE YZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "HEXADECAPOLE ZZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical hexadecapole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "LOWDIN_CHARGES": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "atomic charges using Lowdin's method",
        "dimensionality": ["n_atoms"],
        "type": "float",
        "units": ""
    },
    "MAYER_INDICES": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "Mayer's bond indices",
        "dimensionality": ["triangular", "n_atoms", "n_atoms"],
        "type": "float",
        "units": ""
    },
    "MULLIKEN_CHARGES": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "atomic charges using Mulliken's method",
        "dimensionality": ["n_atoms"],
        "type": "float",
        "units": ""
    },
    "NUCLEAR REPULSION ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE XXX": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE XXY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE XXZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE XYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE XYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE XZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE YYY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE YYZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE YZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "OCTUPOLE ZZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical octupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "ONE-ELECTRON ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "the one-electron energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "PCM POLARIZATION ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "the PCM polarization energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "QUADRUPOLE XX": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical quadrupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "QUADRUPOLE XY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical quadrupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "QUADRUPOLE XZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical quadrupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "QUADRUPOLE YY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical quadrupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "QUADRUPOLE YZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical quadrupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "QUADRUPOLE ZZ": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical quadrupole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "SCF DIPOLE X": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical dipole moment from SCF",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "SCF DIPOLE Y": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical dipole moment from SCF",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "SCF DIPOLE Z": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "electrical dipole moment from SCF",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "SCF ITERATIONS": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "number of itereations in the SCF",
        "dimensionality": "scalar",
        "type": "integer",
        "units": ""
    },
    "TWO-ELECTRON ENERGY": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "the two-electron energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": ""
    },
    "WIBERG_LOWDIN_INDICES": {
        "calculation": [
            "energy",
            "optimization",
            "thermodynamics",
            "vibrations"
        ],
        "description": "the Wiber-Lowdin bond indices",
        "dimensionality": ["triangular", "n_atoms", "n_atoms"],
        "type": "float",
        "units": ""
    },
}  # yapf: disable



# -*- coding: utf-8 -*-

"""The graphical part of a Psi4 Energy node"""

import configargparse
import logging
import tkinter as tk
import tkinter.ttk as ttk

import psi4_step
import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkEnergy(seamm.TkNode):

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=120,
        y=20,
        w=200,
        h=50,
        my_logger=logger,
        keyword_metadata=None
    ):
        """Initialize the graphical Tk Psi4 energy step

        Keyword arguments:
        """
        self.results_widgets = []

        # Argument/config parsing
        self.parser = configargparse.ArgParser(
            auto_env_var_prefix='',
            default_config_files=[
                '/etc/seamm/psi4_energy.ini',
                '/etc/seamm/seamm.ini',
                '~/.seamm/psi4_energy.ini',
                '~/.seamm/seamm.ini',
            ]
        )

        self.parser.add_argument(
            '--seamm-configfile',
            is_config_file=True,
            default=None,
            help='a configuration file to override others'
        )

        # Options for this plugin
        self.parser.add_argument(
            "--psi4-tk-energy-log-level",
            default=configargparse.SUPPRESS,
            choices=[
                'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
            ],
            type=lambda string: string.upper(),
            help="the logging level for the Psi4 Tk_energy step"
        )

        self.options, self.unknown = self.parser.parse_known_args()

        # Set the logging level for this module if requested
        if 'psi4_tk_energy_log_level' in self.options:
            my_logger.setLevel(self.options.psi4_tk_energy_log_level)
            my_logger.critical(
                'Set log level to {}'.format(
                    self.options.psi4_tk_energy_log_level
                )
            )

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
            keyword_metadata=keyword_metadata
        )

    def right_click(self, event):
        """Probably need to add our dialog...
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(
        self, title='Edit Psi4 Energy Step', calculation='energy'
    ):
        """Create the dialog!"""
        self.logger.debug('Creating the dialog')
        frame = super().create_dialog(
            title=title, widget='notebook', results_tab=True
        )

        # Create a frame for the calculation control
        self['calculation'] = ttk.LabelFrame(
            frame,
            borderwidth=4,
            relief='sunken',
            text='Calculation',
            labelanchor='n',
            padding=10
        )

        # Create all the widgets
        P = self.node.parameters
        for key in psi4_step.EnergyParameters.parameters:
            if key not in ('results', 'extra keywords', 'create tables'):
                self[key] = P[key].widget(self['calculation'])

        # bindings...
        self['level'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_calculation
        )
        self['level'].config(state='readonly')

        self['method'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_calculation
        )
        self['method'].combobox.bind("<Return>", self.reset_calculation)
        self['method'].combobox.bind("<FocusOut>", self.reset_calculation)

        self['advanced_method'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_calculation
        )
        self['advanced_method'].combobox.bind(
            "<Return>", self.reset_calculation
        )
        self['advanced_method'].combobox.bind(
            "<FocusOut>", self.reset_calculation
        )

        self.setup_results(psi4_step.properties, calculation=calculation)

        # Top level needs to call reset_dialog
        if calculation == 'energy':
            self.reset_dialog()

        self.logger.debug('Finished creating the dialog')

    def reset_dialog(self, widget=None):
        """Layout the widgets as needed for the current state"""

        frame = self['frame']
        for slave in frame.grid_slaves():
            slave.grid_forget()

        self['calculation'].grid(row=0, column=0)
        self.reset_calculation()
        return 1

    def reset_calculation(self, widget=None):
        level = self['level'].get()

        if level == 'recommended':
            long_method = self['method'].get()
            method = psi4_step.methods[long_method]['method']
            functional = self['functional'].get()
            meta = psi4_step.methods[long_method]
        else:
            long_method = self['advanced_method'].get()
            method = psi4_step.methods[long_method]['method']
            functional = self['advanced_functional'].get()
            meta = psi4_step.methods[long_method]

        frame = self['calculation']
        for slave in frame.grid_slaves():
            slave.grid_forget()

        widgets = []
        widgets2 = []
        row = 0
        self['level'].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
        row += 1
        if level == 'recommended':
            self['method'].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self['method'])
            row += 1
            if method == 'dft':
                self['functional'].grid(row=row, column=1, sticky=tk.EW)
                widgets2.append(self['functional'])
                row += 1
            if 'freeze core?' in meta and meta['freeze core?']:
                self['freeze-cores'].grid(row=row, column=1, sticky=tk.EW)
                widgets2.append(self['freeze-cores'])
                row += 1
        else:
            self['advanced_method'].grid(
                row=row, column=0, columnspan=2, sticky=tk.EW
            )
            widgets.append(self['advanced_method'])
            row += 1
            if method == 'dft':
                self['advanced_functional'].grid(
                    row=row, column=1, sticky=tk.EW
                )
                widgets2.append(self['advanced_functional'])
                row += 1
            if 'freeze core?' in meta and meta['freeze core?']:
                self['freeze-cores'].grid(row=row, column=1, sticky=tk.EW)
                widgets2.append(self['freeze-cores'])
                row += 1
        if method == 'dft':
            dispersions = psi4_step.dft_functionals[functional]['dispersion']
            if len(dispersions) > 1:
                w = self['dispersion']
                w.config(values=dispersions)
                if w.get() not in dispersions:
                    w.value(dispersions[1])
                w.grid(row=row, column=1, sticky=tk.W)
                widgets2.append(self['dispersion'])
                row += 1
                sw.align_labels(widgets2)
            frame.columnconfigure(0, minsize=30)
        self['spin-restricted'].grid(
            row=row, column=0, columnspan=2, sticky=tk.EW
        )
        widgets.append(self['spin-restricted'])
        row += 1
        # self['calculate gradients'].grid(row=row, column=0, sticky=tk.EW)
        # widgets.append(self['calculate gradients'])
        # row += 1
        sw.align_labels(widgets)

        return row



# -*- coding: utf-8 -*-

"""Setup and run Psi4"""

import json
import logging
from pathlib import Path

import psi4_step
import seamm
import seamm.data
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('psi4')


class Energy(seamm.Node):

    def __init__(self, flowchart=None, title='Energy', extension=None):
        """Initialize the node"""

        logger.debug('Creating Energy {}'.format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.parameters = psi4_step.EnergyParameters()

        self.description = 'A single point energy calculation'

    @property
    def header(self):
        """A printable header for this section of output"""
        return (
            'Step {}: {}'.format(
                '.'.join(str(e) for e in self._id), self.title
            )
        )

    @property
    def version(self):
        """The semantic version of this module.
        """
        return psi4_step.__version__

    @property
    def git_revision(self):
        """The git version of this module.
        """
        return psi4_step.__git_revision__

    def description_text(self, P=None, calculation_type='Single-point energy'):
        """Prepare information about what this node will do
        """

        if not P:
            P = self.parameters.values_to_dict()

        if P['level'] == 'recommended':
            method = P['method']
        else:
            method = P['advanced_method']

        if psi4_step.methods[method]['method'] == 'dft':
            if P['level'] == 'recommended':
                functional = P['functional']
            else:
                functional = P['advanced_functional']
            text = f'{calculation_type} using {method} with an '
            text += f'exchange-correlation potential of {functional}'
            if (
                len(psi4_step.dft_functionals[functional]['dispersion']) > 1
                and P['dispersion'] != 'none'
            ):
                text += f" with the {P['dispersion']} dispersion correction."
            else:
                text += " with no dispersion correction."
        else:
            text = f'{calculation_type} using {method}.'

        # Spin
        if P['spin-restricted'] == 'yes':
            text += " The spin will be restricted to a pure eigenstate."
        elif P['spin-restricted']:
            text += " The spin will not be restricted and may not be a "
            text += "proper eigenstate."
        else:
            text += " Whether the spin will be restricted to a pure "
            text += "eigenstate will be determined by {P['spin-restricted']}"

        return self.header + '\n' + __(text, **P, indent=4 * ' ').__str__()

    def get_input(self, calculation_type='energy'):
        """Get the input for an energy calculation for Psi4"""
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # references = self.parent.references

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = '{:~P}'.format(PP[key])

        self.description = []
        self.description.append(
            __(self.description_text(PP), **PP, indent=self.indent)
        )

        lines = []
        if calculation_type == 'energy':
            lines.append('')
            lines.append('#' * 80)
            lines.append(f'# {self.header}')
            lines.append('#' * 80)

        # Figure out what we are doing!
        if P['level'] == 'recommended':
            method_string = P['method']
        else:
            method_string = P['advanced_method']

        method = psi4_step.methods[method_string]['method']

        lines.append('set scf_type df')
        lines.append('set guess sad')

        spin_restricted = P['spin-restricted'] == 'yes'
        if method == 'dft':
            if spin_restricted:
                lines.append('set reference rks')
            else:
                lines.append('set reference uks')
        else:
            if spin_restricted:
                lines.append('set reference rhf')
            else:
                lines.append('set reference uhf')
        lines.append('')
        if method == 'dft':
            if P['level'] == 'recommended':
                functional_string = P['functional']
            else:
                functional_string = P['advanced_functional']
            functional = psi4_step.dft_functionals[functional_string]['name']
            if len(
                psi4_step.dft_functionals[functional_string]['dispersion']
            ) > 1:
                functional = functional + '-' + P['dispersion']
            lines.append(
                f"E, wfn = {calculation_type}('{functional}', return_wfn=True)"
            )
        else:
            lines.append(
                f"E, wfn = {calculation_type}('{method}', return_wfn=True)"
            )

        # Dump the properties to a json file
        filename = f'@{self._id[-1]}+properties.json'
        lines.append('')
        lines.append('oeprop(')
        lines.append('    wfn,')
        lines.append("    'MULTIPOLE(5)',")
        lines.append("    'ESP_AT_NUCLEI',")
        lines.append("    'MO_EXTENTS',")
        lines.append("    'LOWDIN_CHARGES',")
        lines.append("    'MULLIKEN_CHARGES',")
        lines.append("    'WIBERG_LOWDIN_INDICES',")
        lines.append("    'MAYER_INDICES',")
        lines.append("    'NO_OCCUPATIONS',")
        lines.append("    title='PROP'")
        lines.append(')')
        lines.append('')
        lines.append('variables = wfn.scalar_variables()')
        lines.append('arrays = wfn.array_variables()')
        lines.append('for item in arrays:')
        lines.append(
            '    variables[item] = wfn.array_variable(item).np.tolist()'
        )
        lines.append("variables['E'] = E")
        lines.append('')
        lines.append(f"with open('{filename}', 'w') as fd:")
        lines.append('    json.dump(variables, fd, sort_keys=True, indent=3)')

        return '\n'.join(lines)

    def analyze(self, indent='', data={}, out=[]):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """

        # P = self.parameters.current_values_to_dict(
        #     context=seamm.flowchart_variables._data
        # )

        # Read in the results from json
        directory = Path(self.directory)
        json_file = directory / 'properties.json'
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)

            # Put any requested results into variables or tables
            self.store_results(
                data=data,
                properties=psi4_step.properties,
                results=self.parameters['results'].value,
                create_tables=self.parameters['create tables'].get()
            )

            text = 'The calculated energy is {E:.6f} Ha.'
        else:
            data = {}
            text = (
                '\nThere are no results from Psi4. Perhaps it '
                f'failed? Looking for {str(json_file)}.'
            ),

        printer.normal(__(text, **data, indent=self.indent + 4 * ' '))



# -*- coding: utf-8 -*-

"""Setup and run Psi4"""

import json
import logging
from pathlib import Path

import psi4_step
import seamm
import seamm.data
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('psi4')


class Optimization(psi4_step.Energy):

    def __init__(self, flowchart=None, title='Optimization', extension=None):
        """Initialize the node"""

        logger.debug('Creating Optimization {}'.format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.parameters = psi4_step.OptimizationParameters()

        self.description = 'A geometry optimization'

    def description_text(self, P=None):
        """Prepare information about what this node will do
        """

        if not P:
            P = self.parameters.values_to_dict()

        text = super().description_text(
            P=P, calculation_type='Geometry optimization'
        )

        added = (
            'The geometry optimization will use the {optimization method} '
        )
        if P['max geometry steps'] == 'default':
            added += 'method, using the default maximum number of steps, which'
            added += ' is based on the system size.'
        else:
            added += 'method, with no more than {max geometry steps} steps.'

        if P['geometry convergence'] == 'Custom':
            added += ' The convergence criterion is'
        else:
            added += " The convergence criterion is '{geometry convergence}'."

        if P['recalc hessian'] != 'never':
            added += " The Hessian will be recalculated every {recalc hession}"
            added += " steps. Note that calculating the second derivatives is "
            added += "quite expensive!"

        return text + '\n' + __(added, **P, indent=4 * ' ').__str__()

    def get_input(self, calculation_type='optimize'):
        """Get the input for an optimization calculation for Psi4"""
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # references = self.parent.references

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = '{:~P}'.format(PP[key])

        self.description = []
        self.description.append(
            __(self.description_text(PP), **PP, indent=self.indent)
        )

        lines = []
        lines.append('')
        lines.append('#' * 80)
        lines.append(f'# {self.header}')
        lines.append('#' * 80)
        lines.append('')
        lines.append('set opt_type min')
        opt_method = psi4_step.optimization_methods[P['optimization method']]
        lines.append(f'set step_type {opt_method}')
        max_steps = 50
        lines.append(f'set geom_maxiter {max_steps}')
        if P['geometry convergence'] == 'Custom':
            pass
        else:
            convergence = psi4_step.optimization_convergence[
                P['geometry convergence']]
            lines.append(f'set g_convergence {convergence}')
        lines.append('')

        # Add in the input from the energy part of things
        lines.append(super().get_input(calculation_type=calculation_type))

        return '\n'.join(lines)

    def analyze(self, indent='', data={}, out=[]):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """

        # P = self.parameters.current_values_to_dict(
        #     context=seamm.flowchart_variables._data
        # )

        # Read in the results from json
        directory = Path(self.directory)
        json_file = directory / 'properties.json'
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)

            # Put any requested results into variables or tables
            self.store_results(
                data=data,
                properties=psi4_step.properties,
                results=self.parameters['results'].value,
                create_tables=self.parameters['create tables'].get()
            )

            text = 'The calculated energy is {E:.6f} Ha.'
        else:
            data = {}
            text = (
                '\nThere are no results from Psi4. Perhaps it '
                f'failed? Looking for {str(json_file)}.'
            ),

        printer.normal(__(text, **data, indent=self.indent + 4 * ' '))
